from django.contrib.auth import get_user_model
from django.core import signing
from django.utils import baseconv

__all__ = (
    'UserDoesNotExist', 'UserSigner',
)


class UserDoesNotExist(signing.BadSignature):
    pass


class UserSigner(signing.TimestampSigner):
    """Issue and verify URL safe access tokens for users."""

    def __init__(self, key=None, sep='.', salt=None):
        super().__init__(key=key, sep=sep, salt=salt)

    @staticmethod
    def to_timestamp(value):
        """
        Return URL safe base62 encoded timestamp of given datetime.

        Args:
            value (datetime.datetime): Datetime instance to encode.

        Returns:
            str: URL safe base62 encoded timestamp.

        """
        if value is None:
            return ''
        return baseconv.base62.encode(int(value.timestamp()))

    def sign(self, user):
        """
        Return access token for given user.

        The access token will not include a expiration date, but a timestamp
        when it has been issued.

        Args:
            user (django.contrib.auth.base_user.AbstractBaseUser):
                User object to issue a token for.

        Returns:
            str: URL safe base64 string.

        """
        return super().sign(value=self._make_hash_value(user))

    def _make_hash_value(self, user):
        last_login = self.to_timestamp(user.last_login)
        user_pk = baseconv.base62.encode(user.pk)
        return self.sep.join((user_pk, last_login))

    def unsign(self, value, max_age=None, single_use=True):
        """
        Verify access token and return user, if the token is valid.

        Args:
            value (str): URL safe base64 encoded access token.
            max_age (datetime.timedelta): Maximum age an access token to be valid.
            single_use (bool):
                If ``True``, the same token can only be used once and will be invalid
                the next try.
                If ``False``, the same token can be used multiple times and remains
                valid until expired.
                Default: ``True``
        Returns:
            django.contrib.user.models.BaseUser: Return user object for given
            access token.

        Raises:
            UserDoesNotExist:
                If user object does not exist in the database. This can happen,
                if the user object has been deleted.
            SignatureExpired:
                If the access token has expired or of the user has been logged in
                after the token has been created. The latter prevents users
                from using tokens multiple times.
            django.core.signing.BadSignature:
                If the signature does not match provided data.

        """
        result = super().unsign(value, max_age=max_age)
        user_pk, last_login = result.rsplit(self.sep, 2)
        user_pk = baseconv.base62.decode(user_pk)
        try:
            user = get_user_model()._default_manager.get(pk=user_pk)
        except get_user_model().DoesNotExist as e:
            raise UserDoesNotExist("User with pk=%s does not exist" % user_pk) from e
        else:
            if (single_use and last_login != '' and
                    self.to_timestamp(user.last_login) != last_login):
                raise signing.SignatureExpired(
                    "The access token for %r seems used" % user
                )
            return user



import logging

from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.core.signing import BadSignature, SignatureExpired
from django.urls import reverse

from . import signing
from .signing import UserDoesNotExist

logger = logging.getLogger(__name__)


class MailAuthBackend(ModelBackend):
    signer = signing.UserSigner()

    def authenticate(self, request, token=None):
        max_age = getattr(settings, 'LOGIN_URL_TIMEOUT', 60 * 15)
        single_use = getattr(settings, 'LOGIN_TOKEN_SINGLE_USE', True)

        try:
            user = self.signer.unsign(token, max_age=max_age, single_use=single_use)
        except UserDoesNotExist:
            logger.warning(
                "Valid token for non-existing user. Maybe the user has been deleted.",
                exc_info=True,
            )
        except SignatureExpired:
            logger.warning("Token has expired.", exc_info=True)
        except BadSignature:
            logger.exception('Malicious or corrupted login token received: "%s"', token)
        else:
            if self.user_can_authenticate(user):
                return user

    @classmethod
    def get_token(cls, user):
        return cls.signer.sign(user)

    @staticmethod
    def get_login_url(token):
        return reverse(
            'mailauth:login-token',
            kwargs={'token': token}
        )



from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME, authenticate, login
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.core.exceptions import PermissionDenied
from django.http import response
from django.shortcuts import resolve_url
from django.utils.decorators import method_decorator
from django.views import generic
from django.views.decorators.cache import never_cache

from mailauth import forms

__all__ = (
    'LoginView', 'LoginTokenView',
)


class LoginView(DjangoLoginView):
    """
    Send a login code to the user.

    It doesn't authenticate a user but it is the entry point for the login
    process (login URL).
    """

    form_class = forms.EmailLoginForm
    success_url = getattr(settings, 'LOGIN_REQUESTED_URL', 'mailauth:login-success')
    template_name = 'registration/login.html'
    redirect_field_name = REDIRECT_FIELD_NAME

    def form_valid(self, form):
        form.save()
        return response.HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        return resolve_url(self.success_url)

    def get_initial(self):
        return {
            self.redirect_field_name: self.request.GET.get(self.redirect_field_name),
            **super().get_initial(),
        }


INTERNAL_LOGIN_URL_TOKEN = 'login-token'


class LoginTokenView(DjangoLoginView):
    """Authenticate a user via a access token."""

    redirect_field_name = REDIRECT_FIELD_NAME

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        token = kwargs['token']

        user = authenticate(request, token=token)
        if user is None:
            raise PermissionDenied
        else:
            login(self.request, user=user)
            # Remove token from the HTTP Referer header
            self.request.path.replace(token, INTERNAL_LOGIN_URL_TOKEN)

        return response.HttpResponseRedirect(self.get_success_url())

    def post(self, request, *args, **kwargs):
        return self.http_method_not_allowed(request, *args, **kwargs)


class LoginRequestedView(generic.TemplateView):
    template_name = 'registration/login_requested.html'
